"""Decorator to automatically parse request items from the Azure function.

Copyright (c) 2023, Felix Geilert
"""

from azure.functions import HttpRequest

from functown.utils import BaseDecorator
from .handler import RequestArgHandler


class ArgsHandler(BaseDecorator):
    """Decorator to automatically parse HttpRequest parameters.

    This will add an `args` parameter of type `RequestArgHandler` to the inner function
    signature.
    """

    def __init__(self, **kwargs):
        super().__init__(None, added_kw=["args"], **kwargs)

    def run(self, func, *args, **kwargs):
        req: HttpRequest = self._get("req", 0, *args, **kwargs)
        reqhandler = RequestArgHandler(req)
        kwargs["args"] = reqhandler

        return func(*args, **kwargs)



"""
Helper Class to decode JWT Tokens and verify the claims.

Note: This is currently required for python functions as they have no claim principal

Copyright 2022 - Felix Geilert
"""

import base64
from dataclasses import dataclass
import logging
from time import time
from typing import Dict, Any, Union, Set, List, Optional

import requests
from jose import jwt
from cryptography.hazmat.primitives.asymmetric.rsa import RSAPublicNumbers
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization

from functown.errors import TokenError


BEARER = "bearer "


@dataclass
class Token:
    """Token object that contains the claims of the token.

    Args:
        name (str): Passed username in the token
        oid (str): Unique ID of the user
        scopes (set): Set of access scopes for the user from the token
        local_mode (bool): Defines if function runs in local mode
    """

    user_id: str
    scopes: Set[str]
    verified: bool = False
    claims: Optional[Dict[str, Any]] = None
    _local: bool = False

    @property
    def local(self) -> bool:
        """Returns if the function is running in local mode."""
        return self._local


def _ensure_bytes(key: Union[str, bytes]) -> bytes:
    """Make sure the key is bytes"""
    if isinstance(key, str):
        key = key.encode("utf-8")
    return key


def _decode_value(val):
    """Decodes a base64 encoded string."""
    decoded = base64.urlsafe_b64decode(_ensure_bytes(val) + b"==")
    return int.from_bytes(decoded, "big")


def decode_token(
    headers,
    issuer_url: Optional[str] = None,
    audience: Optional[str] = None,
    verify: bool = True,
    logger: Optional[logging.Logger] = None,
) -> Token:
    """Verifies the request headers and returns a list of claims.

    Note that this will not verify the token as this should be done by azure.
    This also requires that the JWT token contains at least the following claims:
    - `oid`: The unique ID of the user
    - `scp`: The access scopes of the user
    - `exp`: The expiration time of the token

    Args:
        headers: req.header information to retrieve the token
        issuer_url (str): The issuer url to retrieve the public key. Only required if
            `verify` is set to `True`.
        audience (str): The audience to verify the token against. Only required if
            `verify` is set to `True`.
        verify (bool): Whether to verify the token signature against the issuer.
            Defaults to `True`.

    Returns:
        name (str): Passed username in the token
        oid (str): Unique ID of the user
        scopes (set): Set of access scopes for the user from the token
        local_mode (bool): Defines if function runs in local mode

    Raises:
        TokenError: If the token could not be parsed (this is a user facing error)
        ValueError: If the issuer url is not provided
        IOError: If the issuer data could not be retrieved
    """
    # retrieve logger
    logger = logger or logging

    # get token
    hdict = dict(headers)
    token = hdict.get("authorization", None)
    min_len = len(BEARER)

    # validate if the token exists
    if token and len(token) > min_len:
        token = token[min_len:]
    else:
        raise TokenError("JWT Token could not be parsed")

    # check if in localmode (for azure functions)
    host = hdict.get("host", None)
    local_mode = host and host.split(":")[0].lower() == "localhost"

    # check if token should be verified
    if verify is True:
        # check if issuer url is provided
        if issuer_url is None:
            raise ValueError("Issuer URL is required for verification")
        if audience is None:
            logger.warning("Audience is not provided for verification")

        # parse headers from the JWT token
        header_data = jwt.get_unverified_header(token)

        # retrieve issuer data
        issuer_res = requests.request("get", issuer_url)
        if issuer_res.status_code != 200:
            raise IOError(f"Unable to obtain issuer data from {issuer_url}")
        issuer = issuer_res.json()

        # retrieve keys
        key_res = requests.request("get", issuer["jwks_uri"])
        if key_res.status_code != 200:
            raise IOError("Unable to obtain issuer key data")
        keys = key_res.json()

        # verify that kid matches
        sig = None
        for key in keys["keys"]:
            if header_data["kid"] == key["kid"]:
                sig = key
                break
        if not sig:
            raise TokenError("Token signature does not match")

        # retrieve pub key
        pk = (
            RSAPublicNumbers(n=_decode_value(sig["n"]), e=_decode_value(sig["e"]))
            .public_key(default_backend())
            .public_bytes(
                encoding=serialization.Encoding.PEM,
                format=serialization.PublicFormat.SubjectPublicKeyInfo,
            )
        )

        # retrieve the algo from the header and decode
        claims = jwt.decode(
            token,
            key=pk,
            algorithms=[header_data["alg"]],
            audience=audience,
            issuer=issuer["issuer"],
        )
    else:
        logger.warning("JWT token is not verified!")
        claims = jwt.get_unverified_claims(token)

    # verify the claims
    if not local_mode:
        # retrieve azure passed data
        user_id = hdict.get("x-ms-client-principal-id", None)
        user_name = hdict.get("x-ms-client-principal-name", None)

        # compare
        if (user_id and user_id != claims["oid"]) or (
            user_name and user_name != claims.get("name", user_name)
        ):
            raise TokenError("Provided user information does not match")
    else:
        logger.info("Executing local mode (avoid user check)")

    # verify if token is expired
    cur_time = time()
    if claims["exp"] < cur_time:
        logger.error(
            f"Current time is {cur_time} but token is expired at {claims['exp']} "
            f"(diff: {claims['exp'] - cur_time})"
        )
        raise TokenError("Token is expired")

    # further parse the claims
    scopes = set(claims["scp"].split(" "))

    # pass on the data
    return Token(claims["oid"], scopes, verify, claims, local_mode)


def verify_user(
    req,
    scopes: Optional[List[str]] = None,
    issuer_url: Optional[str] = None,
    audience: Optional[str] = None,
    verify: bool = False,
    debug: bool = False,
    logger: Optional[logging.Logger] = None,
) -> Token:
    """Verifies the user send in the request

    Args:
        req: HttpRequest that should be verified
        scopes: List of Scopes (strings) or single scope
        verify: Defines if the JWT token should be verified with the provider

    Returns:
        user: User name in the Token
        user_id: User Id in the token
        scopes: Scopes from the token
        local: If the request was send from localhost
    """
    logger = logger or logging
    # verify the token
    try:
        token = decode_token(
            req.headers, issuer_url, audience, verify=verify, logger=logger
        )
    except TokenError as ex:
        raise TokenError(ex.msg, 401)
    except Exception as ex:
        logger.error(ex)
        raise TokenError("Unable to parse token", 500)

    # verify scopes
    if scopes:
        # make sure scopes are list
        if not isinstance(scopes, (list, tuple)):
            scopes = [scopes]

        # iterate scopes
        for scp in scopes:
            if scp not in token.scopes:
                if debug is True:
                    raise TokenError(
                        f"Missing scope {scp} in current token: {token.scopes}", 401
                    )
                else:
                    raise TokenError(
                        "Missing authorization scope for current access", 401
                    )

    # provide data
    return token



"""Helper Functions to retrieve flags from Azure Environments.

Copyright (c) 2023, Felix Geilert
"""

import os
from distutils.util import strtobool
from typing import Optional


def get_config(name: str, default: Optional[str] = None) -> Optional[str]:
    """Retrieves a configuration value from the Azure Function Environment.

    Args:
        name (str): Name of the configuration value.
        default (Optional[str], optional): Default value if the configuration value is
            not set. Defaults to `None`.
    """
    return os.environ.get(name, default)


def get_flag(name: str, default: bool = False) -> bool:
    """Retrieves a flag from the Azure Function Environment.

    Args:
        name (str): Name of the flag.
        default (bool, optional): Default value if the flag is not set. Defaults to
            `False`.
    """
    return bool(strtobool(os.environ.get(name, str(default))))



from .callbacks import (
    filter_debug,
    create_filter_ids,
    create_filter_keywords,
    modify_system_info,
)

from .base import InsightsDecorator

from .metrics import (
    MetricType,
    Metric,
    MetricSpec,
    MetricTimeValue,
    MetricHandler,
    MetricUseMode,
)

from .logger_decorator import InsightsLogs
from .event_decorator import InsightsEvents
from .tracer_decorator import InsightsTracer, TracerObject
from .metric_decorator import InsightsMetrics
from .all_decorator import insights



"""Parsing of protobuf messages.

Copyright (c) 2023, Felix Geilert
"""

from typing import Any, Dict, Tuple, Union

from azure.functions import HttpRequest
from functown.args import RequestArgHandler, HeaderEnum, ContentTypes
from functown.errors import RequestError
from google.protobuf import json_format

from .base import SerializationDecorator, DeserializationDecorator


class ProtobufResponse(SerializationDecorator):
    """Provides a protobuf serialized response for an Azure Function.

    Args:
        func (Callable): The function to decorate.
        pb_class (Any): The protobuf class to use for serialization. If provided, will
            perform a hard type check on the response. Defaults to `None`.
        headers (Dict[str, str]): The headers to add to the response.
        status_code (int): The status code of the response.
        allow_json (bool): Whether to allow dict or list objects as input. In this case
            the object will automatically be converted into the regarding protobuf
            class. (Note that this strictly requires pb_class to be provided).
            Defaults to `None` (True if pb_class provided).

    Example:
        >>> @ProtobufResponse
        ... def main(req: HttpRequest) -> Dict[str, str]:
        ...     pb = pb2.Example()
        ...     return pb

        >>> @ProtobufResponse(headers={"X-My-Header": "My-Value"})
        ... def main(req: HttpRequest) -> Dict[str, str]:
        ...     pb = pb2.Example()
        ...     return pb
    """

    def __init__(
        self,
        func=None,
        pb_class: Any = None,
        headers: Dict[str, str] = None,
        status_code: int = 200,
        allow_json: bool = None,
        **kwargs,
    ):
        super().__init__(func, headers, status_code, **kwargs)
        self._pb_class = pb_class
        self._allow_json = (
            allow_json if allow_json is not None else pb_class is not None
        )

    def serialize(
        self, req: HttpRequest, res: Any, *args, **kwargs
    ) -> Tuple[Union[bytes, str], str]:
        # check if already serialized
        if isinstance(res, bytes):
            return res, ContentTypes.BINARY

        # perform type check (if requested)
        if self._pb_class is not None:
            # check if json or list and convert
            if self._allow_json is True and isinstance(res, (list, dict)):
                res = json_format.ParseDict(res, self._pb_class())
            elif not isinstance(res, self._pb_class):
                raise ValueError(f"Response is not of type {self._pb_class.__name__}")

        # check for SerializeToString method
        if not hasattr(res, "SerializeToString"):
            raise ValueError("Response does not have a SerializeToString method")

        return res.SerializeToString(), ContentTypes.BINARY


class ProtobufRequest(DeserializationDecorator):
    """Provides a Protobuf deserialized request for an Azure Function.

    This decorator will add a `body` argument to the decorated function. This argument
    will contain the deserialized protobuf object.

    Args:
        func (Callable): The function to decorate.
        pb_class (Any): The protobuf class to use for deserialization.
        enfore_mime (bool): Whether to enforce the mimetype check of the request body.
            Defaults to `True`.
        allow_json (bool): Whether to allow JSON requests. This will automatically parse
            a JSON request into the regarding protobuf object. Defaults to `False`.

    Example:
        >>> @ProtoRequest()
        ... def main(req: HttpRequest, body: Any):
        ...     item = body.item
    """

    def __init__(
        self,
        pb_class: Any,
        enforce_mime: bool = True,
        allow_json: bool = True,
        **kwargs,
    ):
        super().__init__(None, **kwargs)

        # check if ParseFromString method exists
        if not hasattr(pb_class, "ParseFromString"):
            raise ValueError("Provided class does not have a ParseFromString method")

        # set data
        self._enforce = enforce_mime
        self._pb_class = pb_class
        self._allow_json = allow_json

    def deserialize(self, req: HttpRequest, *args, **kwargs) -> Any:
        # validate mimetype
        mime = RequestArgHandler(req).get_header(
            HeaderEnum.CONTENT_TYPE, required=self._enforce
        )
        mime = mime.split(";")[0].lower() if mime is not None else None

        # check for json data
        if mime == ContentTypes.JSON.value.lower() and self._allow_json is True:
            body = req.get_body()
            if isinstance(body, str):
                body = body.encode("utf-8")
            pb_item = self._pb_class()
            json_format.Parse(body, pb_item)
            return pb_item

        # check for hard request
        if self._enforce is True and mime != ContentTypes.BINARY.value.lower():
            raise RequestError(f"Request body must be octet-stream (is {mime}).", 400)

        # retrieve body and decode to string
        body = req.get_body()
        if isinstance(body, str):
            body = body.encode("utf-8")
        elif not isinstance(body, bytes):
            raise RequestError("Request body must be a str or bytes object.", 400)

        # FEAT: enable deeper parsing check
        pb_item = self._pb_class()
        pb_item.ParseFromString(body)
        return pb_item



"""Decorator for tracing function calls inside of Application Insights.

Examples how to use tracing:

```python
# basic trace
with tracer.span("my_span"):
    # do something

# trace with extra data
with tracer.span("my_span", attributes={"key": "value"}):
    # do something
```

For more info on tracing, see:
https://opencensus.io/tracing/

Copyright (c) 2023, Felix Geilert
"""


import logging

from opencensus.ext.azure.trace_exporter import AzureExporter
from opencensus.trace.samplers import ProbabilitySampler
from opencensus.trace.tracer import Tracer

from .base import InsightsDecorator

TracerObject = Tracer


class InsightsTracer(InsightsDecorator):
    """Decorator that provides a sending interface to the Application Insights.

    This will add the `tracer` object to the function call.

    Args:
        instrumentation_key (str): The instrumentation key for the Application Insights.
        sampling_rate (float): The sampling rate for the tracer. Defaults to 1.0.
    """

    def __init__(
        self,
        instrumentation_key: str,
        sampling_rate: float = 1.0,
        **kwargs,
    ):
        super().__init__(instrumentation_key, added_kw=["tracer"], **kwargs)

        self.sampling_rate = sampling_rate

    def run(self, func, *args, **kwargs):
        try:
            # create exporter
            exporter = None
            if self.instrumentation_key is not None:
                exporter = AzureExporter(
                    connection_string=f"InstrumentationKey={self.instrumentation_key}"
                )

            # generate tracer
            tracer = Tracer(
                exporter=exporter,
                sampler=ProbabilitySampler(self.sampling_rate),
            )
            kwargs["tracer"] = tracer
        except Exception as ex:
            logging.error(f"Failed to create Insights Tracer: {ex}")
            raise ex

        return func(*args, **kwargs)



"""Convert Request and Response data using Flatbuffers.

Copyright (c) 2023, Felix Geilert
"""

from typing import Dict, Any, Tuple, Union

from azure.functions import HttpRequest
import flatbuffers  # noqa: F401
from functown.args import ContentTypes, RequestArgHandler, HeaderEnum
from functown.errors import RequestError

from .base import SerializationDecorator, DeserializationDecorator


class FlatbufferResponse(SerializationDecorator):
    """Provides a Flatbuffer serialized response for an Azure Function.

    Args:
        func (Callable): The function to decorate.
        fb_class (Any): The flatbuffer class to use for serialization. If provided, will
            perform a hard type check on the response. Defaults to `None`.
        headers (Dict[str, str]): The headers to add to the response.
        status_code (int): The status code of the response.
        allow_json (bool): Whether to allow dict or list objects as input. In this case
            the object will automatically be converted into the regarding flatbuffer
            class. (Note that this strictly requires fb_class to be provided).
            Defaults to `None` (True if fb_class provided).

    Example:
        >>> @FlatbufferResponse
        ... def main(req: HttpRequest) -> fb.Example:
        ...     fb = fb2.Example()
        ...     return fb
    """

    def __init__(
        self,
        func=None,
        fb_class: Any = None,
        headers: Dict[str, str] = None,
        status_code: int = 200,
        **kwargs,
    ):
        super().__init__(func, headers, status_code, **kwargs)
        self._fb_class = fb_class

    def serialize(
        self, req: HttpRequest, res: Any, *args, **kwargs
    ) -> Tuple[Union[bytes, str], str]:
        # check if already serialized
        if isinstance(res, bytes):
            return res, ContentTypes.BINARY

        # perform type check (if requested)
        if self._fb_class is not None:
            if not isinstance(res, self._fb_class):
                raise ValueError(f"Response is not of type {self._fb_class.__name__}")

        # check for SerializeToString method
        if not hasattr(res, "Output"):
            raise ValueError("Response does not have a Output method")

        return bytes(res.Output()), ContentTypes.BINARY


class FlatbufferRequest(DeserializationDecorator):
    """Provides a Flatbuffer deserialized request for an Azure Function.

    This decorator will add a new `body` argument to the decorated function. This
    argument will contain the deserialized flatbuffer object.

    Args:
        fb_class (Any): The flatbuffer class to use for deserialization.
        enfore_mime (bool): Whether to enforce the mimetype check of the request body.
            Defaults to `True`.
        allow_json (bool): Whether to allow JSON requests. This will automatically parse
            a JSON request into the regarding flatbuffer object. Defaults to `False`.

    Example:
        >>> @FlatbufferRequest()
        >>> def func(req: HttpRequest, data: fb.Example) -> Any:
        >>>     # do something with data
        >>>     return ...
    """

    def __init__(
        self,
        fb_class: Any,
        enforce_mime: bool = True,
        **kwargs,
    ):
        super().__init__(None, **kwargs)

        self._fb_class = fb_class
        self._enforce = enforce_mime

        if self._fb_class is None:
            raise ValueError("fb_class must be set")
        if not hasattr(self._fb_class, "GetRootAs"):
            raise ValueError("fb_class must have a GetRootAs method")

    def deserialize(self, req: HttpRequest, *args, **kwargs) -> Any:
        # validate mimetype
        mime = RequestArgHandler(req).get_header(
            HeaderEnum.CONTENT_TYPE, required=self._enforce
        )
        mime = mime.split(";")[0].lower() if mime is not None else None

        # check for hard request
        if self._enforce is True and mime != ContentTypes.BINARY.value.lower():
            raise RequestError(f"Request body must be octet-stream (is {mime}).", 400)

        # retrieve body and decode to string
        body = req.get_body()
        if isinstance(body, str):
            body = body.encode("utf-8")
        elif not isinstance(body, bytes):
            raise RequestError("Request body must be a str or bytes object.", 400)

        # generate the response object
        item = self._fb_class.GetRootAs(bytearray(body), 0)
        return item



"""Decorator base class.

Used for stackable decorators to avoid boilerplate and clean at the out layer.

Copyright (c) 2023, Felix Geilert
"""


import logging
from inspect import Parameter, Signature, signature
import threading
from typing import Dict, Any, Union, List, Callable, Tuple


# defines type used for ids
IDTYPE = str


class BaseDecorator(object):
    """Base Decorator class.

    This decorator can either be used without init or with init:
    - without init: @BaseDecorator
    - with init: @BaseDecorator()

    Note that the variant with init is advised, as it preserves the signature of the inner function.

    You can disable the non-init variant by removing the func parameter from the init.

    ```python
    class MyDecorator(BaseDecorator):
        def __init__(self, *args, **kwargs):
            super().__init__(None, *args, **kwargs)

        # ...
    ```

    In order to properly mask the signature, each implementation has to pass the addded_kw
    parameter for each added keyword to the function signature.
    """

    # internal counter to keep track of the decorator level per function
    # NOTE: this has to be done on a class level, since each decorator is a new instance
    __decorator_count: Dict[IDTYPE, Tuple[int, IDTYPE]] = {}

    def __init__(
        self,
        func=None,
        added_kw: List[str] = None,
        mask_signature: bool = True,
        **kwargs,
    ):
        # NOTE: this retrieves either only `func` (in non-init case) or decorator parameters
        # check if a function is passed (in case of @BaseDecorator, i.e. non-init)
        self.func = func
        self._is_init = func is None

        # create basic variables
        self.mask_signature = mask_signature
        self.added_kw = added_kw if added_kw else []
        self.__address = None

    @property
    def level(self) -> Union[int, None]:
        """Returns the level of the decorator (i.e. how many other decorators have been called before).

        First decorator is level 0, second is level 1, etc.
        """
        tpl = self.__class__.__decorator_count.get(self.__address, None)
        if tpl is None:
            return None
        return tpl[0] - 1

    @property
    def max_level(self) -> Union[int, None]:
        # retrieve current data
        if self.__address is None:
            return None
        tpl = self.__class__.__decorator_count.get(self.__address, None)
        if tpl is None:
            return None

        # find the max level
        base_id = tpl[1]
        max_lvl = tpl[0]
        for lvl, ref_id in self.__class__.__decorator_count.values():
            if ref_id == base_id and lvl > max_lvl:
                max_lvl = lvl
        return max_lvl - 1

    @property
    def is_first_decorator(self) -> Union[bool, None]:
        """Returns True if this is the first decorator in the stack (i.e. outermost)."""
        if self.__address is None:
            return None

        # get current data
        cur_lvl, base_id = self.__class__.__decorator_count[self.__address]

        # validate if higher data
        for lvl, addr in self.__class__.__decorator_count.values():
            if addr == base_id and lvl > cur_lvl:
                return False
        return True

    @property
    def is_last_decorator(self) -> bool:
        """Returns True if this is the last decorator in the stack (i.e. innermost)."""
        return self.level == 0

    def _get(self, name: str, pos: int = None, *args, **kwargs) -> Union[Any, None]:
        """Retrieves an item either by name or by position."""
        if name in kwargs:
            return kwargs[name]
        if pos is not None and len(args) > pos:
            return args[pos]
        return None

    def _create_logger(
        self, force_clean: bool = False, name="logger", *args, **kwargs
    ) -> logging.Logger:
        """Creates a logger for the function.

        If the logger is already passed, it is returned.

        Args:
            force_clean (bool, optional): If True, the logger is always created. Defaults to False.
            name (str, optional): The name of the logger. Defaults to "logger".

        Returns:
            logging.Logger: The logger.
        """
        logger = self._get(name, None, *args, **kwargs)
        if logger is None or force_clean is True:
            name = self.__class__.__name__
            logging.debug(f"Creating logger for {name}")
            logger = logging.getLogger(name)
        return logger

    def run(self, func, *args, **kwargs):
        """Implement this function with additional parameters as the actual decorator logic.

        In this form, this is just a pass through.

        In order to have proper function of the inner function,
        this should always end with a return func(*args, **kwargs).
        """
        return func(*args, **kwargs)

    def teardown(self, *args, **kwargs):
        """Implement this function to perform teardown logic."""
        pass

    def __modify_sig(self, sig: Signature, kws: List[str]) -> Signature:
        """Removes the given keywords from the signature."""
        kws = [kw for kw in kws if kw in sig.parameters]
        if kws:
            sig = sig.replace(
                parameters=[p for p in sig.parameters.values() if p.name not in kws]
            )

        return sig

    def __id(self, obj) -> IDTYPE:
        """Returns the id of the object."""
        thread_id = threading.current_thread().ident
        return f"{thread_id}-{id(obj)}"
        # return id(obj)

    def __set_base_id(self, func) -> IDTYPE:
        """Returns the id of the function and sets the base id for the decorator."""
        base_id = self.__id(func)
        self.__class__.__decorator_count[base_id] = (0, None)
        return base_id

    def __find_closure(self, func):
        """Returns the id of the first BaseDecorator or None."""
        if func.__closure__ is None:
            return None

        obj = [
            cl.cell_contents
            for cl in func.__closure__
            if cl.cell_contents is not None
            and issubclass(cl.cell_contents.__class__, BaseDecorator)
        ]
        obj = [self.__id(o) for o in obj[:1]]

        return obj[0] if len(obj) > 0 else None

    def __increase_count(self):
        """Executes the increase of the decorator count."""
        base_id, level = None, 0
        self.__address = self.__id(self)
        ref_id = self.__find_closure(self.func)
        if ref_id is None or ref_id not in self.__class__.__decorator_count:
            base_id = self.__set_base_id(self.func)
        else:
            level, base_id = self.__class__.__decorator_count[ref_id]
        self.__class__.__decorator_count[self.__address] = (level + 1, base_id)

    def __call__(self, *args, **kwargs) -> Union[Any, Callable[[Any], Any]]:
        """Call the decorator.

        Note that if the function is not passed to the init (only for @BaseDecorator),
        This function is always passed here

        This function has two call modes:
        - non-init decorator: retrieves the call arguments from the inner function and
            returns function result.
        - init decorator: retrieves only the `func` as call argument and returns an
            updated function object.
        """
        # check the case
        if self._is_init is True:
            # retrieve the function from the arguments
            self.func = self._get("func", 0, *args, **kwargs)

            def execute(*args, **kwargs):
                # execute function
                res = self.run(self.func, *args, **kwargs)
                # execute teardown
                self.teardown(*args, **kwargs)
                # return result
                return res

            # update the reference pointer
            self.__increase_count()

            # clean additional arguments from the signature return for the outer decorator
            if self.mask_signature:
                sig = signature(self.func)
                kws = (
                    []
                    if self.is_first_decorator is False
                    else [
                        kw
                        for kw in sig.parameters
                        if sig.parameters[kw].kind
                        in [Parameter.VAR_KEYWORD, Parameter.VAR_POSITIONAL]
                    ]
                )
                execute.__signature__ = self.__modify_sig(sig, self.added_kw + kws)

            return execute

        # --- code for non-init decorator ---
        # increase reference counters
        self.__increase_count()
        return self.run(self.func, *args, **kwargs)



import logging

from .base import SerializationDecorator, DeserializationDecorator
from .json import JsonResponse, JsonRequest

try:
    from .flatbuf import FlatbufferResponse, FlatbufferRequest
except ImportError:
    logging.warning(
        "Unable to load flatbuffers, please install `functown[flatbuffers]`"
    )

try:
    from .protobuf import ProtobufResponse, ProtobufRequest
except ImportError:
    logging.warning("Unable to load protobuf, please install `functown[protobuf]`")

try:
    from .dataframe import DataFrameResponse, DataFrameRequest, DataFrameFormat
except ImportError:
    logging.warning("Unable to load pandas, please install `functown[pandas]`")



"""Combined Decorator for all insights elements.

Copyright (c) 2023, Felix Geilert
"""

from typing import Callable, List

from opencensus.ext.azure.log_exporter import Envelope

from functown.utils import StackDecorator

from .metrics import MetricSpec
from .event_decorator import InsightsEvents
from .logger_decorator import InsightsLogs
from .metric_decorator import InsightsMetrics
from .tracer_decorator import InsightsTracer


def insights(
    instrumentation_key: str,
    enable_logger: bool = False,
    send_basics: bool = False,
    logger_callback: Callable[[Envelope], bool] = None,
    clean_logger: bool = False,
    enable_events: bool = False,
    event_callback: Callable[[Envelope], bool] = None,
    clean_events: bool = False,
    enable_metrics: bool = False,
    metrics: List[MetricSpec] = None,
    enable_tracer: bool = False,
    sampling_rate: float = 1.0,
    **kwargs,
):
    """Decorator that combines functionality from LogHandler, MetricsHandler,
    TraceHandler and EventHandler.

    Args:
        instrumentation_key (str): The instrumentation key for the Application Insights.
        enable_logger (bool): Whether to enable logging to Application Insights.
            Defaults to `False`.
        send_basics (bool): Whether to send basic information about the function call
            to Application Insights. Defaults to `False`.
        logger_callback (Callable[[Envelope], bool]): A callback function that is called
            after the logger has been called. Defaults to `None`.
        clean_logger (bool): Whether to clean the logger after the function call.
            Defaults to `False`.
        enable_events (bool): Whether to enable events to Application Insights.
            Defaults to `False`.
        event_callback (Callable[[Envelope], bool]): A callback function that is called
            after the event has been called. Defaults to `None`.
        clean_events (bool): Whether to clean the events after the function call.
            Defaults to `False`.
        enable_metrics (bool): Whether to enable metrics to Application Insights.
            Defaults to `False`.
        metrics (List[MetricSpec]): A list of metrics to be logged. Defaults to `None`.
        enable_tracer (bool): Whether to enable tracing to Application Insights.
            Defaults to `False`.
        sampling_rate (float): The sampling rate for the tracer. Defaults to `1.0`.
    """
    decs = []
    if enable_logger:
        logger = InsightsLogs(
            instrumentation_key,
            send_basics=send_basics,
            callback=logger_callback,
            clean_logger=clean_logger,
        )
        decs.append(logger)

    if enable_events:
        events = InsightsEvents(
            instrumentation_key,
            callback=event_callback,
            clean_events=clean_events,
        )
        decs.append(events)

    if enable_metrics:
        metrics = InsightsMetrics(instrumentation_key, metrics=metrics)
        decs.append(metrics)

    if enable_tracer:
        tracer = InsightsTracer(instrumentation_key, sampling_rate=sampling_rate)
        decs.append(tracer)

    return StackDecorator(decs)



"""Converts data from and to dataframes.

Supporting the following formats:
- CSV
- JSON
- Parquet

Copyright (c) 2023, Felix Geilert
"""

from enum import Enum
from io import StringIO, BytesIO
from typing import Dict, Tuple, Union, Optional

from azure.functions import HttpRequest
import pandas as pd

from functown.args import ContentTypes, RequestArgHandler, HeaderEnum
from functown.errors import RequestError

from .base import SerializationDecorator, DeserializationDecorator


class DataFrameFormat(str, Enum):
    """Supported DataFrame formats."""

    CSV = "csv"
    JSON = "json"
    PARQUET = "parquet"


class DataFrameResponse(SerializationDecorator):
    """Provides a DataFrame serialized response for an Azure Function.

    Args:
        func (Callable): The function to decorate.
        format (DataFrameFormat): The format to use for serialization. Defaults to
            `DataFrameFormat.CSV`.
        headers (Dict[str, str]): The headers to add to the response.
        status_code (int): The status code of the response.
        allow_empty (bool): Whether to allow empty dataframes as input. Defaults to
            `True`.

    Example:
        >>> @DataFrameResponse
        ... def main(req: HttpRequest) -> pd.DataFrame:
        ...     return pd.DataFrame(...)
    """

    def __init__(
        self,
        func=None,
        format=DataFrameFormat.CSV,
        headers: Optional[Dict[str, str]] = None,
        status_code: int = 200,
        allow_empty: bool = True,
        **kwargs,
    ):
        super().__init__(func, headers, status_code, **kwargs)

        self._format = format
        self._allow_empty = allow_empty

    def serialize(
        self, req: HttpRequest, res: pd.DataFrame, *args, **kwargs
    ) -> Tuple[Union[bytes, str], str]:
        # check if dataframe is valid
        if not isinstance(res, pd.DataFrame):
            raise ValueError(f"Response is not a valid DataFrame (is {type(res)})")
        if self._allow_empty is False and (res is None or res.empty):
            raise ValueError("Response is empty")
        is_empty = res is None or res.empty

        # serialize dataframe according to format
        if self._format == DataFrameFormat.CSV:
            data = res.to_csv(index=False) if not is_empty else None
            mime = ContentTypes.CSV
        elif self._format == DataFrameFormat.JSON:
            data = res.to_json(orient="records") if not is_empty else None
            mime = ContentTypes.JSON
        elif self._format == DataFrameFormat.PARQUET:
            data = res.to_parquet() if not is_empty else None
            mime = ContentTypes.BINARY
        else:
            raise ValueError(f"Unsupported DataFrame format: {self._format}")

        return data, mime


class DataFrameRequest(DeserializationDecorator):
    """Provides a DataFrame deserialized request for an Azure Function.

    This decorator will add a `df` keyword argument to the decorated function. This
    argument will contain the deserialized DataFrame.

    Args:
        func (Callable): The function to decorate.
        fixed_format (DataFrameFormat): The format to use for deserialization. If
            `None`, the format will be determined from the `Content-Type` header.
            Defaults to `None`.
        enforce_mime (bool): Whether to enforce the `Content-Type` header to match
            the format. Defaults to `False`.
    """

    def __init__(
        self,
        func=None,
        fixed_format: Optional[DataFrameFormat] = None,
        enforce_mime: bool = False,
        **kwargs,
    ):
        super().__init__(func, keyword="df", **kwargs)

        self._fixed_format = fixed_format
        self._enforce = enforce_mime
        self._fixed_mime = {
            DataFrameFormat.CSV: ContentTypes.CSV,
            DataFrameFormat.JSON: ContentTypes.JSON,
            DataFrameFormat.PARQUET: ContentTypes.BINARY,
        }.get(fixed_format, None)

    def deserialize(self, req: HttpRequest, *args, **kwargs) -> pd.DataFrame:
        # validate mime type
        mime = RequestArgHandler(req).get_header(
            HeaderEnum.CONTENT_TYPE, required=self._enforce
        )
        mime = mime.split(";")[0].lower() if mime is not None else None

        # check for hard request
        if self._fixed_mime is not None and mime != self._fixed_mime.value.lower():
            raise RequestError(
                f"Request body must be {self._fixed_mime.value} (is {mime}).", 400
            )
        if self._enforce is True and mime is None:
            raise RequestError("Request body must have a Content-Type header.", 400)

        # normalize body
        body = req.get_body()
        if body is None:
            return None
        # NOTE: using lambdas for lazy evaluation
        if isinstance(body, str):
            body_str, body_bytes = lambda: body, lambda: body.encode("utf-8")
        elif isinstance(body, bytes):
            body_str, body_bytes = lambda: body.decode("utf-8"), lambda: body
        else:
            raise RequestError(f"Unsupported body type: {type(body)}", 400)

        # parse dataframe
        if mime == ContentTypes.CSV.value.lower():
            df = pd.read_csv(StringIO(body_str()))
        elif mime == ContentTypes.JSON.value.lower():
            df = pd.read_json(StringIO(body_str()))
        elif mime == ContentTypes.BINARY.value.lower():
            df = pd.read_parquet(BytesIO(body_bytes()))
        else:
            raise RequestError(f"Unsupported content type: {mime}", 400)

        return df



"""Helper classes for metrics.

Note that the time_series returned from OpenCensus internally do not always reflect
the correct timestamps, but are bucketed. Therefore items can cluster together.

This code should provide the primary abstraction interface for OpenCensus API.
No code outside this file should need to touch the opencensus API directly.

Copyright (c) 2023, Felix Geilert
"""


from dataclasses import dataclass
from datetime import datetime
from enum import Enum
import logging
from typing import Dict, Union, Type, List, Any, Callable

from opencensus.ext.azure import metrics_exporter
from opencensus.stats.measurement_map import MeasurementMap
from opencensus.stats import aggregation as aggregation_module
from opencensus.stats import measure as measure_module
from opencensus.stats import stats as stats_module
from opencensus.stats import view as view_module
from opencensus.tags import tag_map as tag_map_module
from opencensus.metrics.export.metric import Metric as OCMetric
from opencensus.metrics.export.time_series import TimeSeries as OCTimeSeries

from functown.utils.metaclasses import ThreadSafeSingleton


class MetricType(Enum):
    """The type of the metric."""

    COUNTER = 1
    GAUGE = 2
    SUM = 3


class MetricUseMode(Enum):
    """The use mode of the metric."""

    HARD_FAIL = 1
    """Hard Fail in the case of a namespace conflict"""
    SOFT_FAIL = 2
    """Soft Fail in the case of a namespace conflict"""
    OVERWRITE = 3
    """Overwrite in the case of a namespace conflict"""
    REUSE = 4
    """Reuse in the case of a namespace conflict"""


@dataclass
class MetricSpec:
    """A definition of a metric.

    Define namespaces if you have co-dependent metrics that you want to always record
    together.

    Args:
        name (str): The name of the metric.
        description (str): The description of the metric.
        unit (str): The unit of the metric.
        columns (List[str]): The columns of the metric. These will be used to
            log additional data into the metric. Note that metric values aggregations
            happens based on a group by these columns (and timestamps).
        mtype (MetricType): The type of the metric. Defines how data logged is
            aggregated.
        dtype (Type[Union[int, float]]): The data type of the metric. Defaults to int.
        namespace (str): The namespace of the metric. Metrics in the same namespace will
            be recorded together. Note that this might lead to increases in counter
            metrics (as None values are logged that are counted). Per default every
            metric is in its own namespace.
    """

    name: str
    description: str
    unit: str
    columns: List[str]
    mtype: MetricType
    dtype: Type[Union[int, float]] = int
    namespace: str = None
    start_value: Union[int, float, None] = None

    @property
    def used_namespace(self) -> str:
        """The namespace that is used for the metric.

        If no namespace is defined, the name is used.
        """
        if self.namespace is None:
            return self.name
        return self.namespace


@dataclass
class MetricTimeValue:
    """A value for a metric at a specific time.

    Args:
        value (Union[int, float]): The value of the metric.
        timestamp (datetime): The timestamp of the value.
        columns (Dict[str, Any]): The columns of the metric.
    """

    value: Union[int, float]
    timestamp: datetime
    columns: Dict[str, Any]

    def __str__(self):
        return f"{self.value} at {self.timestamp} with {self.columns}"


class Metric:
    """Handler class for metrics that allow to record values

    Args:
        spec (MetricSpec): The specification of the metric.
        vm (stats_module.ViewManager): The view manager.
        map (MeasurementMap): The measurement map (should it be reused).
        add_name_column (bool): Whether to add the name of the metric as a column.
        handler_columns (Dict[str, Any]): Additional columns specified by the handler.
            These will always be added to the metric if set. Defaults to None.
    """

    _NAME_COL = "__name"

    def __init__(
        self,
        spec: MetricSpec,
        vm: stats_module.ViewManager,
        map: MeasurementMap,
        add_name_column: bool = True,
        handler_columns: Dict[str, Any] = None,
    ):
        # store the spec
        self.spec = spec
        self.map = map
        self.add_name_column = add_name_column
        self.required_columns = handler_columns or {}

        # select the measure
        measure = {
            float: measure_module.MeasureFloat,
            int: measure_module.MeasureInt,
        }[spec.dtype]

        # select the aggregation
        agg = {
            MetricType.COUNTER: aggregation_module.CountAggregation,
            MetricType.GAUGE: aggregation_module.LastValueAggregation,
            MetricType.SUM: aggregation_module.SumAggregation,
        }[spec.mtype]

        # generate the columns
        cols = spec.columns or []
        if add_name_column:
            cols = [self._NAME_COL] + cols
        cols += list(self.required_columns.keys())
        # remove duplicates (preserving order)
        cols = list(dict.fromkeys(cols))

        # generate data
        self.measure = measure(spec.name, spec.description, spec.unit)
        self.view = view_module.View(
            self.__view_name,
            spec.description,
            cols,
            self.measure,
            agg() if spec.start_value is None else agg(spec.start_value),
        )
        vm.register_view(self.view)

        # generate the default tag map
        self.tag = self.__build_tag_map({})

    @property
    def __view_name(self):
        return f"{self.spec.name}_view"

    def __build_tag_map(self, columns: Dict[str, Any]) -> tag_map_module.TagMap:
        """Builds a tag map from the columns."""
        tag = tag_map_module.TagMap()
        if self.add_name_column:
            tag.insert(self._NAME_COL, self.spec.name)

        # add the required columns (ordering is important)
        # note that user columns overwrite required columns
        cols = {**self.required_columns, **columns}
        for key, value in cols.items():
            if key not in self.spec.columns:
                logging.warning(
                    f"Key {key} is not a valid column for metric {self.sepc.name}. "
                    "Ignoring."
                )
                continue
            tag.insert(key, value)
        return tag

    def add_default_column(self, key: str, value: Any, persistent: bool = False):
        """Adds a default tag to the metric.

        Note: These tags are only provided when no specific tags are provided.

        Args:
            key (str): The key of the tag.
            value (Any): The value of the tag.
            persistent (bool): Persistent columns are always added to the record,
                even if user-specified columns are provided. Defaults to False.
        """
        # ensure that the key is specified
        if key not in self.spec.columns:
            logging.warning(
                f"Key {key} is not a valid column for this metric. Ingnoring."
            )
            return
        self.tag.insert(key, value)
        if persistent:
            self.required_columns[key] = value

    def _record_measure_only(self, value: Union[float, int]):
        """Records a value for the metric."""
        # generate the tags
        if isinstance(value, float):
            self.map.measure_float_put(self.measure, value)
        elif isinstance(value, int):
            self.map.measure_int_put(self.measure, value)
        else:
            raise ValueError(f"Value {value} is not a valid value.")

    def record(self, value: Union[float, int], columns: Dict[str, Any] = None):
        """Records a value for the metric."""
        # generate the tags
        tag_map = self.tag if columns is None else self.__build_tag_map(columns)

        # measure the value and record it to a feature map
        self._record_measure_only(value)
        self.map.record(tag_map)

    @property
    def current_data(self) -> Union[List[Union[int, float]], None]:
        """Returns the data of the metric.

        Note that this will not return timestamps or columns.
        It is also derived from all combined data.

        Returns:
            List of data. In case of gauge and sum only the last value is returned
            (List of size 1). If no data is available, None is returned.
        """
        # retrieve the overall data
        fts = self.full_time_series

        # check if data is available
        if len(fts) == 0:
            return None

        # check the type
        if self.spec.mtype in (MetricType.GAUGE, MetricType.SUM):
            return [self.full_time_series[-1].value]

        # retrieve the overall data
        return [dp.value for dp in self.full_time_series]

    @property
    def full_time_series(self) -> List[MetricTimeValue]:
        """Retrieves the entire time series of the data.

        Note that each column combination is set into a new time series.

        This structure will combine the regarding time series into a single
        (with column values attached to each object)
        """
        return self.get_time_series()

    def get_time_series(
        self,
        start: datetime = None,
        end: datetime = None,
        columns: Dict[str, Any] = None,
    ) -> List[MetricTimeValue]:
        """Returns the timeseries data of this metric filter to specifications.

        Args:
            start (datetime): The start time of the data. If None, all data is returned.
                Defaults to None.
            end (datetime): The end time of the data. If None, the current time is used.
                Defaults to None.
            columns (Dict[str, Any]): The columns to filter for. If None, all columns
                are returned. Defaults to None.

        Returns:
            List[MetricTimeValue]: The time series data.
        """
        # update the datetimes
        if end is None:
            end = datetime.utcnow()
        if start is None:
            start = datetime.min

        # retrieve the overall data
        data: List[OCMetric] = list(self.map.measure_to_view_map.get_metrics(end))
        # retrieve the time series data
        # NOTE: each new tag configuration will create a new time-series
        data = [d for d in data if d.descriptor.name == self.__view_name]
        if len(data) == 0:
            return []
        series: List[OCTimeSeries] = data[0].time_series

        # filter timeseries (based on label_values)
        if columns is not None:
            vals = ([self.spec.name] if self.add_name_column else []) + [
                columns.get(col, None) for col in self.spec.columns
            ]
            series = [s for s in series if [lv.value for lv in s.label_values] == vals]
        elif self.add_name_column:
            # in any case filter for the name
            series = [s for s in series if s.label_values[0].value == self.spec.name]

        # filter and convert the data
        items = []
        for serie in series:
            vals = {
                col: val.value
                for col, val in zip(
                    ([self._NAME_COL] if self.add_name_column is True else [])
                    + self.spec.columns,
                    serie.label_values,
                )
            }
            items += [
                MetricTimeValue(p.value.value, p.timestamp, vals)
                for p in serie.points
                if start <= p.timestamp <= end
            ]

        # sort data by timestamp
        items.sort(key=lambda x: x.timestamp)
        return items


class MetricHandler(metaclass=ThreadSafeSingleton):
    def __init__(self, add_name_column: bool = True):
        # create internal data
        self._metrics: Dict[str, Metric] = {}
        self._enable_standard_metrics = True
        self._add_name_column = add_name_column
        self._maps: Dict[str, MeasurementMap] = {}

        # create opencensus data
        self._vm = stats_module.stats.view_manager
        self._exporter: metrics_exporter.MetricsExporter = None

    def __getattribute__(self, name):
        try:
            return super().__getattribute__(name)
        except AttributeError:
            try:
                return self._metrics[name]
            except KeyError:
                raise AttributeError(name)

    def __getitem__(self, key):
        # allow ["metric"] retrieval
        return self.get_metric(key)

    def get_metric(self, name: str) -> Union[Metric, None]:
        """Returns the metric with the given name.

        Args:
            name (str): The name of the metric.

        Returns:
            Metric: The metric.
        """
        return self._metrics.get(name, None)

    def create_metrics(
        self,
        specs: List[MetricSpec],
        mode: MetricUseMode = MetricUseMode.HARD_FAIL,
        global_columns: Dict[str, Any] = None,
    ) -> bool:
        """Creates a list of metrics based on the specifications.

        Args:
            specs (List[MetricSpec]): The specifications to create the metrics for.
            mode (MetricUseMode): The mode to use when creating the metrics.
            global_columns (Dict[str, Any]): The global columns to add to all metrics.
                Defaults to None.

        Returns:
            bool: Whether the metrics were created successfully.
        """
        # remove items that are already created
        if mode == MetricUseMode.REUSE:
            dropped = [spec.name for spec in specs if spec.name in self._metrics]
            if len(dropped) > 0:
                logging.info(f"Reusing metrics: {', '.join(dropped)}")
                specs = [spec for spec in specs if spec.name not in self._metrics]

        # check if metric names are valid
        if mode in [MetricUseMode.HARD_FAIL, MetricUseMode.SOFT_FAIL]:
            for spec in specs:
                # check for error
                if spec.name in self._metrics or spec.name in self.__dict__:
                    if mode == MetricUseMode.HARD_FAIL:
                        raise ValueError(
                            f"Metric {spec.name} generates a namespace conflict."
                        )
                    return False
        else:
            dropped = [spec.name for spec in specs if spec.name in self.__dict__]
            if len(dropped) > 0:
                logging.info(f"Dropping metrics (illegal names): {', '.join(dropped)}")
                specs = [spec for spec in specs if spec.name not in self.__dict__]

        # find all namespaces and create maps
        namespaces = set([spec.used_namespace for spec in specs])
        for ns in namespaces:
            if ns not in self._maps:
                self._maps[ns] = stats_module.stats.stats_recorder.new_measurement_map()

        # only create metrics if all are valid
        for spec in specs:
            self._metrics[spec.name] = Metric(
                spec,
                self._vm,
                self._maps[spec.used_namespace],
                add_name_column=self._add_name_column,
                handler_columns=global_columns,
            )

        return True

    def record(self, values: Dict[str, Any], columns: Dict[str, Any] = None):
        """Records the given values to the metrics.

        In this case the values for all metrics are recorded under the same tag-map

        NOTE: This will ignore tags set by individual metrics. (Including default tags).

        Args:
            values (Dict[str, Any]): The values to record.
            columns (Dict[str, Any]): The columns to record the values under.
        """
        # generate tag_map
        tag_map = tag_map_module.TagMap()
        if self._add_name_column:
            tag_map.insert(self._NAME_COL, "global")
        for key, value in (columns or {}).items():
            if key == self._NAME_COL:
                raise ValueError(f"Column {key} is reserved.")
            tag_map.insert(key, value)

        # measure the values
        namespaces = []
        for name, value in values.items():
            if name not in self._metrics:
                raise ValueError(f"Metric {name} does not exist.")
            self._metrics[name]._record_measure_only(value)
            namespaces.append(self._metrics[name].spec.used_namespace)

        # complete recording
        for ns in set(namespaces):
            self._maps[ns].record(tag_map)

    @property
    def is_insights_enabled(self) -> bool:
        """Returns whether the exporter is connected to Application Insights."""
        return self._exporter is not None

    def connect_insights(
        self,
        instrumentation_key: str,
        callback: Callable[[metrics_exporter.Envelope], bool],
        enable_standard_metrics: bool = None,
        flush_sec: float = 15,
    ):
        """Connects the metrics to Azure Insights.

        Args:
            instrumentation_key (str): The instrumentation key to connect to.
            callback (Callable[[metrics_exporter.Envelope], bool]): The callback to
                filter the metrics.
            enable_standard_metrics (bool, optional): Whether to enable standard metrics.
                Defaults to None.
            flush_sec (float, optional): The interval to flush the metrics.
                Defaults to 15.
        """
        # update the metrics
        metrics = enable_standard_metrics
        if metrics is None:
            metrics = self._enable_standard_metrics
        self._enable_standard_metrics = metrics

        # generate exporter
        self._exporter = metrics_exporter.new_metrics_exporter(
            enable_standard_metrics=metrics,
            connection_string=f"InstrumentationKey={instrumentation_key}",
            export_interval=flush_sec,
        )

        # check for filter
        if callback is not None:
            self._exporter.add_telemetry_processor(callback)

        self._vm.register_exporter(self._exporter)

    def shutdown(self):
        """Flushes remaining data."""
        if self._exporter is not None:
            self._exporter.shutdown()



"""
Bunch of Handler Code for Arguments

Copyright (c) 2023, Felix Geilert
"""


from distutils.util import strtobool
from enum import Enum
import logging
from typing import Any, Optional, Union, List, Callable

from azure.functions import HttpRequest

from functown.errors import ArgError


# create enum str class with different header names
class HeaderEnum(str, Enum):
    """List of Headers to be used in typed way"""

    AUTHORIZATION = "Authorization"
    CONTENT_TYPE = "Content-Type"
    ACCEPT = "Accept"
    HOST = "Host"
    USER_AGENT = "User-Agent"
    X_FORWARDED_FOR = "X-Forwarded-For"
    X_FORWARDED_PROTO = "X-Forwarded-Proto"
    X_FORWARDED_HOST = "X-Forwarded-Host"
    X_FORWARDED_PORT = "X-Forwarded-Port"
    X_FORWARDED_PREFIX = "X-Forwarded-Prefix"
    X_FORWARDED_PATH = "X-Forwarded-Path"
    X_FORWARDED_QUERY = "X-Forwarded-Query"
    X_FORWARDED_METHOD = "X-Forwarded-Method"
    X_FORWARDED_SCHEME = "X-Forwarded-Scheme"
    X_FORWARDED_PROTOCOL = "X-Forwarded-Protocol"
    X_FORWARDED_URI = "X-Forwarded-Uri"
    X_FORWARDED_URL = "X-Forwarded-Url"


class ContentTypes(str, Enum):
    """List of Content Types to be used in typed way"""

    JSON = "application/json"
    XML = "application/xml"
    PARQUET = "application/parquet"
    CSV = "text/csv"
    HTML = "text/html"
    TEXT = "text/plain"
    FORM = "application/x-www-form-urlencoded"
    MULTIPART = "multipart/form-data"
    BINARY = "application/octet-stream"


class RequestArgHandler:
    def __init__(self, req: HttpRequest):
        self.req = req

        # create a header-map
        if req is None:
            logging.warning("Request is None, no data will be available")
            self._headers = {}
        else:
            headers = list(req.headers.keys())
            self._headers = dict(zip([h.lower() for h in headers], headers))

    def _convert(
        self,
        name: str,
        arg: str,
        required: bool = False,
        map_fct: Callable[[str], Any] = None,
        allowed: Optional[List[Any]] = None,
        list_map: Callable[[Any], List[Any]] = None,
        default: Optional[Any] = None,
    ) -> Optional[Union[Any, List[Any]]]:
        # check if required
        if required and arg is None:
            raise ArgError(f"Argument {name} is not set, but required!", 400)

        # execute the mapping function (if set and arg is present)
        if map_fct is not None and arg is not None:
            if isinstance(map_fct, str):
                if map_fct == "bool":
                    if not isinstance(arg, bool):
                        arg = bool(strtobool(str(arg)))
                else:
                    arg = getattr(arg, map_fct)()
            else:
                arg = map_fct(arg)

        # check for default
        if arg is None and default is not None:
            arg = default

        # checks if in list
        if allowed and arg:
            # check for mapping
            if list_map and isinstance(list_map, str):
                ls_arg = getattr(arg, list_map)()
            else:
                ls_arg = list_map(arg) if list_map else arg

            # check if value found
            if ls_arg not in allowed:
                raise ArgError(
                    f"Argument {name} should be one of {allowed} but got {arg}", 400
                )

        return arg

    def _parse_body(self, name: str) -> Optional[Any]:
        arg = None
        try:
            body = self.req.get_json()
            arg = body[name] if name in body else None
        except Exception:
            logging.warning("Could not parse request body to json")
        return arg

    def get_body(
        self,
        name,
        required=False,
        map_fct=None,
        allowed=None,
        list_map=None,
        default=None,
    ) -> Optional[Union[str, Any, List[Any]]]:
        """Parses a value from the body of the request

        Args:
            name (str): Name of the argument
            required (bool, optional): If the argument is required. Defaults to False.
            map_fct (function, optional): Function to map the value. Defaults to None.
            allowed (list, optional): List of allowed values. Defaults to None.
            list_map (function, optional): Function to map the list. Defaults to None.
            default (any, optional): Default value if not set. Defaults to None.
        """
        arg = self._parse_body(name)
        return self._convert(name, arg, required, map_fct, allowed, list_map, default)

    def get_query(
        self,
        name: str,
        required: bool = False,
        map_fct: Callable[[str], Any] = None,
        allowed: Optional[List[Any]] = None,
        list_map: Callable[[Any], List[Any]] = None,
        default: Optional[Any] = None,
    ) -> Optional[Union[str, Any, List[Any]]]:
        """Parses a value from the query string of the request"""
        arg = self.req.params.get(name)
        return self._convert(name, arg, required, map_fct, allowed, list_map, default)

    def get_route(
        self,
        name: str,
        required: bool = False,
        map_fct: Callable[[str], Any] = None,
        allowed: Optional[List[Any]] = None,
        list_map: Callable[[Any], List[Any]] = None,
        default: Optional[Any] = None,
    ) -> Optional[Union[str, Any, List[Any]]]:
        """Parses a value from the route parameters of the request"""
        arg = self.req.route_params.get(name)
        return self._convert(name, arg, required, map_fct, allowed, list_map, default)

    def get_body_query(
        self,
        name: str,
        required: bool = False,
        map_fct: Callable[[str], Any] = None,
        allowed: Optional[List[Any]] = None,
        list_map: Callable[[Any], List[Any]] = None,
        default: Optional[Any] = None,
    ) -> Optional[Union[str, Any, List[Any]]]:
        """Parses a value from the body or query string of the request"""
        arg = self.req.params.get(name)
        if not arg and self.req.get_body():
            arg = self._parse_body(name)
        return self._convert(name, arg, required, map_fct, allowed, list_map, default)

    def get_form(
        self,
        name: str,
        required: bool = False,
        map_fct: Callable[[str], Any] = None,
        allowed: Optional[List[Any]] = None,
        list_map: Callable[[Any], List[Any]] = None,
        default: Optional[Any] = None,
    ) -> Optional[Union[str, Any, List[Any]]]:
        arg = None
        try:
            arg = self.req.form[name]
        except Exception as ex:
            logging.warning(f"Form value {name} could not be parsed: {ex}")
        return self._convert(name, arg, required, map_fct, allowed, list_map, default)

    def get_file(self, name: str, required: bool = False) -> Optional[Any]:
        """Parses a file from the form data of the request"""
        file = None

        # retrieve data
        try:
            file = self.req.files[name]
        except Exception as ex:
            logging.warning(f"File {name} could not be parsed: {ex}", 400)

        # check if provided
        if required and not file:
            raise ArgError(f"File {name} could not be parsed", 400)

        return file

    def get_header(
        self,
        name: Union[str, HeaderEnum],
        required: bool = False,
        map_fct: Callable[[str], Any] = None,
        allowed: Optional[List[Any]] = None,
        list_map: Callable[[Any], List[Any]] = None,
        default: Optional[Any] = None,
    ) -> Optional[Union[str, Any, List[Any]]]:
        """Parses a header from the request

        Note: You can use `HeaderEnum` to get the correct header name
        """
        # preprocess name
        if isinstance(name, HeaderEnum):
            name = name.value
        name = self._headers.get(name.lower(), name)

        # check if provided
        arg = self.req.headers.get(name)

        if arg is None:
            if required:
                raise ArgError(f"Header {name} is required", 400)
            return None
        return self._convert(name, arg, required, map_fct, allowed, list_map, default)



import os
from setuptools import setup
from setuptools import find_packages
from functown import __version__
import pathlib
from glob import glob


__status__ = "Package"
__copyright__ = "Copyright 2023"
__license__ = "MIT License"
__author__ = "Felix Geilert"


this_directory = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(this_directory, "README.md"), encoding="utf-8") as f:
    long_description = f.read()


def versions_in_requirements(file):
    lines = file.read().splitlines()
    versions = [line for line in lines if not line.isspace() and "--" not in line]
    return list(versions)


HERE = pathlib.Path(__file__).parent


def load_req(extra: str = None, all: bool = False):
    if all is True:
        reqs_files = glob(str(HERE / "requirements*.txt"))
        reqs = []
        for file in reqs_files:
            with open(file) as f:
                reqs += versions_in_requirements(f)

        # make unique
        reqs = list(set(reqs))
        return reqs

    # load regular requirements
    file = "requirements.txt"
    if extra:
        file = f"requirements-{extra}.txt"
    with open(HERE / file) as f:
        required_list = versions_in_requirements(f)
    return required_list


setup(
    name="functown",
    version=__version__,
    description="Various Helper Tools to make working with azure functions easier",
    long_description=long_description,
    long_description_content_type="text/markdown",
    keywords="azure functions",
    url="https://github.com/felixnext/python-functown",
    download_url="https://github.com/felixnext/python-functown/releases/tag/v0.1.0-alpha",
    author="Felix Geilert",
    license="MIT License",
    packages=find_packages(
        include=["functown", "functown.*"],
    ),
    install_requires=load_req(),
    setup_requires=["pytest-runner", "flake8"],
    tests_require=load_req(all=True),
    extras_require={
        "flatbuffer": load_req("flatbuffer"),
        "protobuf": load_req("protobuf"),
        "insights": load_req("insights"),
        "jwt": load_req("jwt"),
        "pandas": load_req("pandas"),
    },
    include_package_data=True,
    zip_safe=False,
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
    ],
)



"""Range of custom errors required in functown.

Copyright (c) 2023, Felix Geilert
"""


class RequestError(Exception):
    """Basic Error for Azure Functions.

    Args:
        msg (str): Error message
        code (int, optional): Error code. Used to generate HttpResponse object.
            Defaults to 500.
    """

    def __init__(self, msg: str, code: int = None, *args: object) -> None:
        super().__init__(*args)
        self.msg = msg
        self.code = 500 if None else code

    def __str__(self) -> str:
        return f"Error ({self.code}): {self.msg}"


class HandlerError(RequestError):
    """Error within one of the decorator functions."""

    pass


class TokenError(RequestError):
    """Errors related to checking token.

    Usually used in the `auth` namespace
    """

    pass


class ArgError(RequestError):
    """Error related to parsing arguments."""

    pass



from typing import List
import logging

from functown.utils import BaseDecorator


class InsightsDecorator(BaseDecorator):
    """Base Class for Application insights decorators."""

    def __init__(self, instrumentation_key: str, added_kw: List[str], **kwargs):
        super().__init__(None, added_kw=added_kw, **kwargs)

        if instrumentation_key is None:
            logging.warning(
                "No instrumentation key provided. "
                "No data will be sent to Application Insights "
                f"({self.__class__.__name__})."
            )

        self.instrumentation_key = instrumentation_key



"""Decorator to automatically parse JWT Tokens from the request of the azure function.

Copyright (c) 2023, Felix Geilert
"""

import logging
from typing import List, Optional

from azure.functions import HttpRequest

from functown.utils import BaseDecorator
from functown.errors import TokenError
from .jwt import verify_user


class AuthHandler(BaseDecorator):
    """Decorator to automatically parse a JWT Token from the HttpRequest parameters.

    This will add an `token` parameter of type `Token` to the inner function signature.

    Args:
        scopes (list): List of scopes that are required for the token
        issuer_url (str): URL of the issuer of the token
        audience (str): Audience of the token
        verify (bool): Whether to verify the token signature against the issuer.
            Defaults to `True`.
        auto_disable_verify (bool): Whether to automatically disable token verification
            if no issuer_url is provided. Defaults to `True`.
        debug (bool): Whether to provide addditional info in error messages for
            debugging. Defaults to `False`.
    """

    def __init__(
        self,
        scopes: Optional[List[str]] = None,
        issuer_url: Optional[str] = None,
        audience: Optional[str] = None,
        verify: bool = True,
        auto_disable_verify: bool = True,
        debug: bool = False,
        **kwargs,
    ):
        super().__init__(None, added_kw=["token"], **kwargs)

        self.scopes = scopes
        self.issuer_url = issuer_url
        self.audience = audience
        if verify and not issuer_url:
            if auto_disable_verify:
                logging.warning(
                    "No issuer_url provided, but verify is True. "
                    "Disabling token verification."
                )
                verify = False
            else:
                raise ValueError("No issuer_url provided, but verify is True.")
        self.verify = verify
        self.debug = debug

    def run(self, func, *args, **kwargs):
        # retrieve request
        req: HttpRequest = self._get("req", 0, *args, **kwargs)

        # get logger
        logger = self._get("logger", None, *args, **kwargs) or logging

        # parse token
        try:
            token = verify_user(
                req,
                scopes=self.scopes,
                issuer_url=self.issuer_url,
                audience=self.audience,
                verify=self.verify,
                logger=logger,
            )
        except TokenError as ex:
            if self.debug:
                raise ex
            raise TokenError("Token validation failed", 401)
        except Exception as ex:
            msg = "Internal error during verification"
            if self.debug:
                msg += f": {ex}"
            raise TokenError(msg, 500)

        kwargs["token"] = token

        return func(*args, **kwargs)



"""Provides Serialization and Deserialization Decorator objects.

Copyright (c) 2023, Felix Geilert
"""

from abc import abstractmethod
from typing import Any, Dict, Tuple, Union

from azure.functions import HttpResponse, HttpRequest

from functown.args import ContentTypes
from functown.utils import BaseDecorator


class SerializationDecorator(BaseDecorator):
    """Provides a serialized response for an Azure Function.

    Derived classes should implement the `serialize` method, which retrieves the
    result of inner function execution as well as remaining function parameters
    (including the request).

    Args:
    """

    def __init__(
        self,
        func=None,
        headers: Dict[str, str] = None,
        status_code: int = 200,
        **kwargs
    ):
        super().__init__(func, **kwargs)

        self._headers = headers
        self._code = status_code

    @abstractmethod
    def serialize(
        self, req: HttpRequest, res: Any, *args, **kwargs
    ) -> Tuple[Union[bytes, str], str]:
        """Serializes the result of the function.

        Args:
            req (HttpRequest): The request object.
            res (Any): The result of the function.

        Returns:
            data (bytes | str): The serialized result.
            mime (str): The mimetype of the serialized result.
        """
        raise NotImplementedError

    def run(self, func, *args, **kwargs):
        # execute inner function
        res = func(*args, **kwargs)

        # get request object
        req = self._get("req", 0, *args, **kwargs)
        if "req" in kwargs:
            del kwargs["req"]
        else:
            args = args[1:]

        # serialize result
        data, mtype = self.serialize(req, res, *args, **kwargs)
        if isinstance(mtype, ContentTypes):
            mtype = mtype.value
        return HttpResponse(
            data, status_code=self._code, headers=self._headers, mimetype=mtype
        )


class DeserializationDecorator(BaseDecorator):
    """Provides a deserialized request body for an Azure Function.

    This will add a `body` parameter to the function signature, which contains the
    deserialized request body.
    """

    def __init__(self, func=None, keyword="body", **kwargs):
        super().__init__(func, [keyword], **kwargs)
        self._kw = keyword

    @abstractmethod
    def deserialize(self, req: HttpRequest, *args, **kwargs) -> Any:
        """Deserializes the request body.

        Args:
            req (HttpRequest): The request object.

        Returns:
            data (Any): The deserialized request body.
        """
        raise NotImplementedError

    def run(self, func, *args, **kwargs):
        # retrieve the request
        req = self._get("req", 0, *args, **kwargs)
        if "req" in kwargs:
            del kwargs["req"]
        else:
            args = args[1:]

        # deserialize request body
        body = self.deserialize(req, *args, **kwargs)
        kwargs[self._kw] = body

        # execute inner function
        return func(req=req, *args, **kwargs)
